VisualThemes = AceLibrary("AceAddon-2.0"):new("AceEvent-2.0", "AceDB-2.0", "AceConsole-2.0", "AceDebug-2.0", "AceHook-2.1", "FuBarPlugin-2.0")
local VisualThemes = VisualThemes
local AceOO = AceLibrary("AceOO-2.0")

local Waterfall = AceLibrary("Waterfall-1.0")

local L = AceLibrary("AceLocale-2.2"):new("VisualThemes")

do
	local rev = tonumber(("$Revision: 48603 $"):match("%d+"))
	VisualThemes.version = "r" .. rev
	VisualThemes.revision = rev
end

VisualThemes.hasIcon = "Interface\\Icons\\INV_Misc_Gem_01"
VisualThemes.blizzardTooltip = true
VisualThemes.hideWithoutStandby = true
VisualThemes.hasNoColor = true
VisualThemes.independentProfile = true

VisualThemes:SetDebugging(false)
VisualThemes:RegisterDB("VisualThemesDB")

local IsEnabled, IsInitialized = false, false

function VisualThemes:IsEnabled()
	return IsEnabled
end

function VisualThemes:IsDisabled()
	return not IsEnabled
end

local options = {
	handler = VisualThemes,
	type = "group",
	args = {
		ThemeOptions = {
			type="group",
			name=L["Theme Options"],
			desc=L["Configure Theme options."], 
			args = {},
		},
	},
}

function VisualThemes:OnInitialize()
	IsInitialized = true
	self:RegisterChatCommand('/VisualThemes', '/VT', {
		type = 'execute',
		func = function()
			Waterfall:Open("VisualThemes")
		end
	})
	 --since i'm not using the options table for the slash command the options table never gets "injected"
	AceLibrary("AceConsole-2.0"):InjectAceOptionsTable(self, options)

	for name, theme in self:IterateThemes() do
		theme:OnInitialize()
	end
end

VisualThemes.OnMenuRequest = options

Waterfall:Register("VisualThemes", "aceOptions", options, "title", L["VisualThemes Configuration"], "treeLevels", 5);

local CallRegisteredAddOns
function VisualThemes:OnEnable(first)
	IsEnabled = true
	--UpdateAddOnMemoryUsage()
	--self:Debug("%s's memory usage is %d", self.name, GetAddOnMemoryUsage(self.name))
	for name, theme in self:IterateThemes() do
		theme:OnEnable(first)
	end

	self:ThemeBlizzardFrames()
	self:PrepareBlizzardDynamicFrames()
	self:RegisterEvent("ADDON_LOADED")
	CallRegisteredAddOns(first)

	--UpdateAddOnMemoryUsage()
	--self:Debug("%s's memory usage is %d", self.name, GetAddOnMemoryUsage(self.name))



	self:UpdateOptionsTable()
end

function VisualThemes:UpdateOptionsTable()
	local themeOptions = options.args.ThemeOptions
	for name, theme in self:IterateThemes() do
		themeOptions.args[theme.Name] = theme.options
	end
end

function VisualThemes:OnClick()
	Waterfall:Open("VisualThemes")
end

do
	local classes = {}
	function VisualThemes:RegisterClass(className, class)
		classes[className] = class
	end

	function VisualThemes:New(className, ...)
		if classes[className] then
			return classes[className]:new(...)
		end
	end
	getmetatable(VisualThemes).__call = VisualThemes.New
end

local ITheme
function VisualThemes:RegisterInterface(interfaceName, interface)
	--check for certain Interfaces important to the core
	if interfaceName == "ITheme" then
		ITheme = interface
	end
end

do
	local defaultTheme
	function VisualThemes:GetDefaultTheme()
		return defaultTheme
	end
	function VisualThemes:HasDefaultTheme()
		return defaultTheme ~= nil
	end
	function VisualThemes:SetDefaultTheme(theme)
		if AceOO.inherits(theme, ITheme) then
			defaultTheme = theme
		end
	end
end

do
	local themes = {}
	function VisualThemes:RegisterTheme(themeName, theme)
		if AceOO.inherits(theme, ITheme) and not themes[themeName] then
			themes[themeName] = theme
			if not self:HasDefaultTheme() then
				self:SetDefaultTheme(theme)
			end
			if IsInitialized then
				theme:OnInitialize()	
			end
			if IsEnabled then
				theme:OnEnable()
				self:UpdateOptionsTable()
			end
			
		end
	end
	function VisualThemes:IterateThemes()
		return pairs(themes)
	end
	function VisualThemes:GetTheme(themeName)
		return themes[themeName]
	end
	function VisualThemes:HasTheme(themeName)
		return themes[themeName] ~= nil
	end
end



do
	local Theme
	local themed = {}
	function VisualThemes:OnDisable()
		IsEnabled = false
		for name, theme in self:IterateThemes() do
			theme:OnDisable()
		end
		for frame in pairs(themed) do
			themed[frame] = nil
		end
	end
	local ThemeChildren = function(theme, depth, ...)
		for i=1, select("#", ...) do
			Theme(select(i, ...), theme, depth > 1 and depth - 1 or nil)
		end
	end
	local function IsSupported(frame)
		local type = frame:GetObjectType()
		return type ~= "Font" and type ~= "FontInstance"
	end
	local masterIgnoreFrameList = {
		MacroFrame = L["causes tainting issues, thus it won't allow macro changes in combat"],
		SpellBookFrame = L["causes tainting issues, thus it open in combat"],
	}
	local function IsValid(frame, theme)
		if type(frame) ~= "table" or not frame.GetObjectType then
			VisualThemes:Print(L["%q cannot be themed, because it's not a frame."], frame ~= nil and frame or "<nil>")
			return false
		end
		if themed[frame] then
			VisualThemes:Print(L["%q cannot be themed, because it's already themed."], frame)
			return false
		end
		if frame:IsProtected() then
			VisualThemes:Print(L["%q cannot be themed, because it's a protected frame."], frame)
			return false
		end
		local name = frame:GetName() or ""
		if masterIgnoreFrameList[name] then
			VisualThemes:Print(L["%q cannot be themed, because: %q."], frame, masterIgnoreFrameList[name] )
			return false
		end
		local isCompat, reason = theme:IsCompatible(name)
		if not isCompat then
			VisualThemes:Print(L["%q cannot be themed, because %s: %q."], frame, theme.Name, reason)
			return false
		end
		if not IsSupported(frame) then
			VisualThemes:Print(L["%q cannot be themed, because it's not a supported type."], frame)
			return false
		end
		return true
	end
	Theme = function(frame, theme, children)
		if IsValid(frame, theme) then
			theme:Theme(frame)
			
			themed[frame] = theme

			if type(children) == "number" and children > 0 then
				return ThemeChildren(theme, children - 1, frame:GetChildren())
			end
		end
	end
	local GetFrame
	do
		local _G = getfenv(0)
		GetFrame = function(name)
			if type(name) == "string" then
				return _G[name]
			end
			return name
		end
	end
	function VisualThemes:Theme(frame, theme, childrenDepth)
		frame = GetFrame(frame)
		if type(theme) == "number" and type(childrenDepth) ~= "number" then
			childrenDepth = theme
			theme = nil
		end
		return Theme(frame, theme and AceOO.inherits(theme, ITheme) and theme or self:GetDefaultTheme(), childrenDepth)
	end
	function VisualThemes:Detheme(frame)
		if themed[frame] then
			themed[frame]:Detheme(frame)
		end
	end
	function VisualThemes:IsThemed(frame, theme)
		if theme and AceOO.inherits(theme, ITheme) then
			return themed[frame] == theme
		end
		return themed[frame] ~= nil
	end
	function VisualThemes:GetFrameTheme(frame)
		return themed[frame]
	end
end

do
	local addonsToTheme, addonDepths, addonFrames = {}, {}, {}
	local function IsAddOnLoadable(addon)
		return select(5, GetAddOnInfo(addon)) ~= nil
	end
	--watches for an addon to be loaded and themes its frames immediately
	function VisualThemes:RegisterAddOn(addon, theme, childrenDepth, ...)
		self:Debug("Registering LoD AddOn for themeing [%q %q %q %q]", addon, theme, childrenDepth, ... ~= nil)
		if type(addon) ~= "string" then return end
		local numFrames = select("#", ...)
		if numFrames <= 0 then return end
		for i=1, numFrames do
			if type(select(i, ...)) ~= "string" then
				return
			end
		end

		if not addonsToTheme[addon] and IsAddOnLoadable(addon) then
			addonsToTheme[addon] = theme or true
			addonDepths[addon] = childrenDepth
			if numFrames > 1 then
				addonFrames[addon] = {...}
			else
				addonFrames[addon] = ...
			end
			if IsAddOnLoaded(addon) then
				for i=1, select("#", ...) do
					self:Theme(select(i, ...), theme, childrenDepth)
				end
			end
		end
	end
	local addonFunc = {}
	CallRegisteredAddOns = function(first)
		for addon, func in pairs(addonFunc) do
			func(self, first)
		end
	end
	--a function is executed when an addon is loaded (or if it's already loaded) 
	--that allows the function to watch for dynamic frames or whatever it needs to do
	--the function is called everytime OnEnable is called on VisualThemes to rehook and reregister events if needed
	function VisualThemes:RegisterAddOnFunc(addon, func)
		if type(addon) ~= "string" then return end
		if type(func) ~= "function" then return end
		if not addonFunc[addon] and IsAddOnLoadable(addon) then
			if IsAddOnLoaded(addon) then
				func(self)
			else
				addonFunc[addon] = func
			end
		end
	end
	function VisualThemes:ADDON_LOADED(addon)
		self:Debug("%q loaded, do we care: %s", addon, addonsToTheme[addon] ~= nil or addonFunc[addon] ~= nil)
		if addonsToTheme[addon] then
			if type(addonFrames[addon]) == "table" then
				for k, frame in ipairs(addonFrames[addon]) do
					if _G[frame] then
						self:Theme(frame, addonDepths[addon] or 0)
					else
						self:Debug("LoD AddOn %q failed to provide frame %q to theme", addon, frame)
					end
				end
			else
				local frame = addonFrames[addon]
				if _G[frame] then
					self:Theme(frame, addonDepths[addon] or 0)
				else
					self:Debug("LoD AddOn %q failed to provide frame %q to theme", addon, frame)
				end
			end
		end
		if addonFunc[addon] then
			addonFunc[addon](self)
		end
	end
end


function VisualThemes:OnTooltipUpdate()
	GameTooltip:AddLine(L["Visual Themes - %s"]:format(self.version))
	GameTooltip:AddLine(L["|cffffff00Left-Click|r to change settings by the GUI."], 0.2, 1, 0.2)
	GameTooltip:AddLine(L["|cffffff00Right-Click|r to change settings by the dropdown."], 0.2, 1, 0.2)
end